@@ -1,14 +1,19 @@
 #This file is part of Tryton.  The COPYRIGHT file at the top level of
 #this repository contains the full copyright notices and license terms.
+import datetime
+import operator
 from decimal import Decimal
-from functools import reduce, partial
-from trytond.model import Workflow, ModelView, ModelSQL, fields
-from trytond.backend import TableHandler
+from functools import partial
+from sql import Column
+from sql.operators import Concat
+
+from trytond.model import Workflow, Model, ModelView, ModelSQL, fields
+from trytond import backend
 from trytond.pyson import In, Eval, Not, Equal, If, Get, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-__all__ = ['Move']
+__all__ = ['StockMixin', 'Move']
 
 STATES = {
     'readonly': In(Eval('state'), ['cancel', 'assigned', 'done']),
@@ -16,6 +21,120 @@ STATES = {
 DEPENDS = ['state']
 
 
+class StockMixin:
+    '''Mixin class with helper to setup stock quantity field.'''
+
+    @classmethod
+    def _quantity_context(cls, name):
+        pool = Pool()
+        Date = pool.get('ir.date')
+
+        context = Transaction().context
+        new_context = {}
+        if name == 'quantity':
+            if (context.get('stock_date_end')
+                    and context['stock_date_end'] > Date.today()):
+                new_context['stock_date_end'] = Date.today()
+        elif name == 'forecast_quantity':
+            new_context['forecast'] = True
+            if not context.get('stock_date_end'):
+                new_context['stock_date_end'] = datetime.date.max
+        return new_context
+
+    @classmethod
+    def _get_quantity(cls, records, name, location_ids, products=None,
+            grouping=('product',), position=-1):
+        """
+        Compute for each record the stock quantity in the default uom of the
+        product.
+
+        location_ids is the list of IDs of locations to take account to compute
+            the stock. It can't be empty.
+        products restrict the stock computation to the this products (more
+            efficient), so it should be the products related to records.
+            If it is None all products are used.
+        grouping defines how stock moves are grouped.
+        position defines which field of grouping corresponds to the record
+            whose quantity is computed.
+
+        Return a dictionary with records id as key and quantity as value.
+        """
+        pool = Pool()
+        Product = pool.get('product.product')
+
+        record_ids = [r.id for r in records]
+        quantities = dict.fromkeys(record_ids, 0.0)
+        if not location_ids:
+            return quantities
+
+        product_ids = products and [p.id for p in products] or None
+
+        with Transaction().set_context(cls._quantity_context(name)):
+            pbl = Product.products_by_location(location_ids=location_ids,
+                product_ids=product_ids, with_childs=True,
+                grouping=grouping)
+
+        for key, quantity in pbl.iteritems():
+            # pbl could return None in some keys
+            if (key[position] is not None and
+                    key[position] in quantities):
+                quantities[key[position]] += quantity
+        return quantities
+
+    @classmethod
+    def _search_quantity(cls, name, location_ids, domain=None,
+            grouping=('product',), position=-1):
+        """
+        Compute the domain to filter records which validates the domain over
+        quantity field.
+
+        location_ids is the list of IDs of locations to take account to compute
+            the stock. It can't be empty.
+        grouping defines how stock moves are grouped.
+        position defines which field of grouping corresponds to the record
+            whose quantity is computed.
+        """
+        pool = Pool()
+        Product = pool.get('product.product')
+
+        if not location_ids or not domain:
+            return []
+
+        def _search_quantity_eval_domain(line, domain):
+            operator_funcs = {
+                '=': operator.eq,
+                '>=': operator.ge,
+                '>': operator.gt,
+                '<=': operator.le,
+                '<': operator.lt,
+                '!=': operator.ne,
+                'in': lambda v, l: v in l,
+                'not in': lambda v, l: v not in l,
+                }
+
+            field, op, operand = domain
+            value = line.get(field)
+            return operator_funcs[op](value, operand)
+
+        with Transaction().set_context(cls._quantity_context(name)):
+            pbl = Product.products_by_location(
+                location_ids=location_ids,
+                with_childs=True, grouping=grouping)
+
+        processed_lines = []
+        for key, quantity in pbl.iteritems():
+            # pbl could return None in some keys
+            if key[position] is not None:
+                processed_lines.append({
+                        'record_id': key[position],
+                        name: quantity,
+                        })
+
+        record_ids = [line['record_id'] for line in processed_lines
+            if _search_quantity_eval_domain(line, domain)]
+        return [('id', 'in', record_ids)]
+
+
 class Move(Workflow, ModelSQL, ModelView):
     "Stock Move"
     __name__ = 'stock.move'
@@ -75,7 +194,7 @@ class Move(Workflow, ModelSQL, ModelView):
             },
         domain=[
             ('id', If(In('company', Eval('context', {})), '=', '!='),
-                Get(Eval('context', {}), 'company', 0)),
+                Eval('context', {}).get('company', -1)),
             ],
         depends=['state'])
     unit_price = fields.Numeric('Unit Price', digits=(16, 4),
@@ -100,6 +219,13 @@ class Move(Workflow, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(Move, cls).__setup__()
+        cls._deny_modify_assigned = set(['product', 'uom', 'quantity',
+            'from_location', 'to_location', 'company', 'unit_price',
+            'currency'])
+        cls._deny_modify_done_cancel = (cls._deny_modify_assigned |
+            set(['planned_date', 'effective_date', 'state']))
+        cls._allow_modify_closed_period = set()
+
         cls._sql_constraints += [
             ('check_move_qty_pos',
                 'CHECK(quantity >= 0.0)', 'Move quantity must be positive'),
@@ -121,8 +247,10 @@ class Move(Workflow, ModelSQL, ModelView):
                 'it is not in draft or cancelled state.'),
             'period_closed': ('You can not modify move "%(move)s" because '
                 'period "%(period)s" is closed.'),
-            'modify_assigned_done_cancel': ('You can not modify stock move "%s"'
-                'because it is in "Assigned", "Done" or "Cancel" state.'),
+            'modify_assigned': ('You can not modify stock move "%s" because '
+                'it is in "Assigned" state.'),
+            'modify_done_cancel': ('You can not modify stock move "%s" '
+                'because it is in "Done" or "Cancel" state.'),
             })
         cls._transitions |= set((
                 ('draft', 'assigned'),
@@ -149,7 +277,10 @@ class Move(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
+
         # Migration from 1.2: packing renamed into shipment
         table = TableHandler(cursor, cls, module_name)
         table.drop_constraint('check_packing')
@@ -177,10 +308,10 @@ class Move(Workflow, ModelSQL, ModelView):
                 for move in moves:
                     internal_quantity = cls._get_internal_quantity(
                             move.quantity, move.uom, move.product)
-                    cursor.execute(
-                        'UPDATE "' + cls._table + '" '
-                        'SET internal_quantity = %s '
-                        'WHERE id = %s', (internal_quantity, move.id))
+                    cursor.execute(*sql_table.update(
+                            columns=[sql_table.internal_quantity],
+                            values=[internal_quantity],
+                            where=sql_table.id == move.id))
             table = TableHandler(cursor, cls, module_name)
             table.not_null_action('internal_quantity', action='add')
 
@@ -199,9 +330,11 @@ class Move(Workflow, ModelSQL, ModelView):
             }
         for column, model in shipments.iteritems():
             if table.column_exist(column):
-                cursor.execute('UPDATE "' + cls._table + '" '
-                    'SET shipment = \'' + model + ',\' || "' + column + '" '
-                    'WHERE "' + column + '" IS NOT NULL')
+                cursor.execute(*sql_table.update(
+                        columns=[sql_table.shipment],
+                        values=[Concat(model + ',',
+                                Column(sql_table, column))],
+                        where=Column(sql_table, column) != None))
                 table.drop_column(column)
 
         # Add index on create_date
@@ -314,9 +447,9 @@ class Move(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def get_shipment(cls):
-        Model = Pool().get('ir.model')
+        IrModel = Pool().get('ir.model')
         models = cls._get_shipment()
-        models = Model.search([
+        models = IrModel.search([
                 ('model', 'in', models),
                 ])
         return [(None, '')] + [(m.model, m.name) for m in models]
@@ -328,17 +461,16 @@ class Move(Workflow, ModelSQL, ModelView):
 
     @classmethod
     def get_origin(cls):
-        Model = Pool().get('ir.model')
+        IrModel = Pool().get('ir.model')
         models = cls._get_origin()
-        models = Model.search([
+        models = IrModel.search([
                 ('model', 'in', models),
                 ])
         return [(None, '')] + [(m.model, m.name) for m in models]
 
-    @classmethod
-    def validate(cls, moves):
-        super(Move, cls).validate(moves)
-        cls.check_period_closed(moves)
+    @property
+    def origin_name(self):
+        return self.origin.rec_name if self.origin else None
 
     @classmethod
     def check_period_closed(cls, moves):
@@ -496,24 +628,26 @@ class Move(Workflow, ModelSQL, ModelView):
             internal_quantity = cls._get_internal_quantity(vals['quantity'],
                 uom, product)
             vals['internal_quantity'] = internal_quantity
-        return super(Move, cls).create(vlist)
+        moves = super(Move, cls).create(vlist)
+        cls.check_period_closed(moves)
+        return moves
 
     @classmethod
     def write(cls, moves, vals):
-        if reduce(lambda x, y: x or y in vals, ('product', 'uom', 'quantity',
-                'from_location', 'to_location', 'company', 'unit_price',
-                'currency'), False):
+        vals_set = set(vals)
+        if cls._deny_modify_assigned & vals_set:
             for move in moves:
-                if move.state in ('assigned', 'done', 'cancel'):
-                    cls.raise_user_error('modify_assigned_done_cancel',
-                        (move.rec_name,))
-        if reduce(lambda x, y: x or y in vals,
-                ('planned_date', 'effective_date', 'state'), False):
+                if move.state == 'assigned':
+                    cls.raise_user_error('modify_assigned', (move.rec_name,))
+        if cls._deny_modify_done_cancel & vals_set:
             for move in moves:
                 if move.state in ('done', 'cancel'):
-                    cls.raise_user_error('modify_assigned_done_cancel',
+                    cls.raise_user_error('modify_done_cancel',
                         (move.rec_name,))
 
+        if any(f not in cls._allow_modify_closed_period for f in vals):
+            cls.check_period_closed(moves)
+
         super(Move, cls).write(moves, vals)
 
         for move in moves:
@@ -557,7 +691,7 @@ class Move(Workflow, ModelSQL, ModelView):
         return to_pick
 
     @classmethod
-    def assign_try(cls, moves):
+    def assign_try(cls, moves, grouping=('product',)):
         '''
         Try to assign moves.
         It will split the moves to assign as much possible.
@@ -579,7 +713,17 @@ class Move(Workflow, ModelSQL, ModelView):
                 stock_assign=True):
             pbl = Product.products_by_location(
                 location_ids=[l.id for l in locations],
-                product_ids=[m.product.id for m in moves])
+                product_ids=[m.product.id for m in moves],
+                grouping=grouping)
+
+        def get_key(location):
+            key = (location.id,)
+            for field in grouping:
+                value = getattr(move, field)
+                if isinstance(value, Model):
+                    value = value.id
+                key += (value,)
+            return key
 
         success = True
         for move in moves:
@@ -591,11 +735,11 @@ class Move(Workflow, ModelSQL, ModelView):
                     ('parent', 'child_of', [move.from_location.id]),
                     ])
             for location in childs:
-                if (location.id, move.product.id) in pbl:
+                key = get_key(location)
+                if key in pbl:
                     location_qties[location] = Uom.compute_qty(
-                            move.product.default_uom,
-                            pbl[(location.id, move.product.id)], move.uom,
-                            round=False)
+                        move.product.default_uom, pbl[key], move.uom,
+                        round=False)
 
             to_pick = move.pick_product(location_qties)
 
@@ -626,10 +770,7 @@ class Move(Workflow, ModelSQL, ModelView):
                 qty_default_uom = Uom.compute_qty(move.uom, qty,
                         move.product.default_uom, round=False)
 
-                pbl[(from_location.id, move.product.id)] = (
-                    pbl.get((from_location.id, move.product.id), 0.0)
-                    - qty_default_uom)
-                pbl[(to_location.id, move.product.id)] = (
-                    pbl.get((to_location.id, move.product.id), 0.0)
-                    + qty_default_uom)
+                from_key, to_key = get_key(from_location), get_key(to_location)
+                pbl[from_key] = pbl.get(from_key, 0.0) - qty_default_uom
+                pbl[to_key] = pbl.get(to_key, 0.0) + qty_default_uom
         return success
